#ifdef HAS_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include <glib/gmacros.h>
#include "rb-gamma.h"

struct _RbGammaPrivate
{
  double gamma;

  GdkColor *colour_100;
  GdkColor *colour_50;

  GdkGC *filled_gc;
  GdkGC *stippled_gc;
};

/* Properties */
enum {
	PROP_0,
	PROP_COLOUR,
	PROP_GAMMA
};

static gpointer parent_class = NULL;

/* TODO: adapt on size of square */
static const char check_bits[] = {0x02, 0x01};

/* TODO: adapt on size of square */
#define NUMBER_OF_BLOCKS 4

/**
 * Take a colour and a gamma correction, and produce a new colour for
 * 50% intensity.
 */
static GdkColor*
gamma_correct (const GdkColor *full, const double gamma)
{
  GdkColor *c = gdk_color_copy (full);
  c->red = CLAMP (c->red * pow (0.5, 1.0/gamma) - 1, 0, 65535);
  c->green = CLAMP (c->green * pow (0.5, 1.0/gamma) - 1, 0, 65535);
  c->blue = CLAMP (c->blue * pow (0.5, 1.0/gamma) - 1, 0, 65535);
  return c;
}

static void
realize_cb (GtkWidget *widget, gpointer data)
{
  RbGammaPrivate *priv;
  GdkColormap *colourmap;
  GdkGC *gc;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (RB_IS_GAMMA (widget));
  priv = RB_GAMMA (widget)->_priv;

  colourmap = gdk_window_get_colormap (widget->window);

  priv->colour_50 = gamma_correct (priv->colour_100, priv->gamma);

  gdk_colormap_alloc_color (colourmap, priv->colour_100, FALSE, TRUE);
  gdk_colormap_alloc_color (colourmap, priv->colour_50, FALSE, TRUE);
  /* TODO: check return value, and maybe pass FALSE to best_match */

  gc = gdk_gc_new (GDK_DRAWABLE (widget->window));
  gdk_gc_set_foreground (gc, priv->colour_50);
  priv->filled_gc = gc;

  gc = gdk_gc_new (GDK_DRAWABLE (widget->window));
  gdk_gc_set_fill (gc, GDK_OPAQUE_STIPPLED);
  gdk_gc_set_foreground (gc, priv->colour_100);
  gdk_gc_set_background (gc, &(widget->style->black));
  gdk_gc_set_stipple (gc, gdk_bitmap_create_from_data (widget->window, check_bits, 2, 2));
  priv->stippled_gc = gc;
}

static void
size_allocate_cb (GtkWidget *widget, GtkAllocation *allocation)
{
  /* Enforce a square gamma grid */
  allocation->width = allocation->height = MIN (allocation->width, allocation->height);
}

static gboolean
expose_event_cb (GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
  RbGammaPrivate *priv;
  int x, y, cell_size;
  gboolean stippled = FALSE;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (RB_IS_GAMMA (widget), FALSE);
  priv = RB_GAMMA (widget)->_priv;
  cell_size = (widget->allocation.width / NUMBER_OF_BLOCKS);

  for (x = 0; x < NUMBER_OF_BLOCKS; ++x) {
    for (y = 0; y < NUMBER_OF_BLOCKS; ++y) {
      gdk_draw_rectangle (widget->window, stippled ? priv->stippled_gc : priv->filled_gc, TRUE, x * cell_size, y * cell_size, cell_size, cell_size);
      stippled = !stippled;
    }
    /* Need this as we have an even number of cells */
    stippled = !stippled;
  }
  return TRUE;
}

static void
rb_gamma_finalize (GObject * gobject)
{
  RbGamma *gamma = RB_GAMMA (gobject);
  g_free (gamma->_priv);
  
  if (G_OBJECT_CLASS (parent_class)->finalize != NULL)
    (*G_OBJECT_CLASS (parent_class)->finalize) (gobject);
}

static void
rb_gamma_set_property (GObject      *object,
                       guint         prop_id,
                       const GValue *value,
                       GParamSpec   *pspec)
{
  RbGamma *gamma = RB_GAMMA (object);
  RbGammaPrivate *priv = gamma->_priv;
  switch (prop_id) {
  case PROP_COLOUR:
    priv->colour_100 = g_value_get_pointer (value);
    break;
  case PROP_GAMMA:
    priv->gamma = g_value_get_double (value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
rb_gamma_get_property (GObject    *object,
                       guint       prop_id,
                       GValue     *value,
                       GParamSpec *pspec)
{
  RbGamma *gamma = RB_GAMMA (object);
  RbGammaPrivate *priv = gamma->_priv;
  switch (prop_id) {
  case PROP_COLOUR:
    g_value_set_pointer (value, priv->colour_100);
    break;
  case PROP_GAMMA:
    g_value_set_double (value, priv->gamma);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
rb_gamma_class_init (RbGammaClass * klass)
{
  GObjectClass *object_class;
  parent_class = g_type_class_peek_parent (klass);

  object_class = (GObjectClass *)klass;
  object_class->finalize = rb_gamma_finalize;
  object_class->set_property = rb_gamma_set_property;
  object_class->get_property = rb_gamma_get_property;

  /* Properties */
  g_object_class_install_property (object_class, PROP_COLOUR,
                                   g_param_spec_pointer ("colour", NULL, NULL,
                                                         G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

  g_object_class_install_property (object_class, PROP_GAMMA,
                                   g_param_spec_double ("gamma", NULL, NULL,
                                                       0.0, 10.0, 1.0,
                                                       G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
}

/* TODO: make stay square */

static void
rb_gamma_init (RbGamma * gamma)
{
  gamma->_priv = g_new0 (RbGammaPrivate, 1);
  gtk_widget_set_size_request (GTK_WIDGET (gamma), 100, 100);
  g_signal_connect (gamma, "realize", G_CALLBACK (realize_cb), NULL);
  g_signal_connect (gamma, "size-allocate", G_CALLBACK (size_allocate_cb), NULL);
  g_signal_connect (gamma, "expose-event", G_CALLBACK (expose_event_cb), NULL);
}

GType
rb_gamma_get_type (void)
{
  static GType type = 0;
  
  if (!type)
    {
      GTypeInfo info = {
        sizeof (RbGammaClass),
        NULL,
        NULL,
        (GClassInitFunc) rb_gamma_class_init,
        NULL,
        NULL,
        sizeof (RbGamma),
        0,
        (GInstanceInitFunc) rb_gamma_init
      };
      
      type = g_type_register_static (GTK_TYPE_DRAWING_AREA, "RbGamma", &info, 0);
    }
  return type;
}

/* For now RbGamma will copy the colour */
GtkWidget *
rb_gamma_new (GdkColor colour, double gamma)
{
  GtkWidget *w;
  w = GTK_WIDGET (g_object_new (RB_TYPE_GAMMA, "colour", gdk_color_copy (&colour), "gamma", gamma, NULL));
  return w;
}
